function [Y,B] = ffpML(X,varargin)
% FFPML --- Fast Fixed-Point algorithm for maximum likelihood estimation.
% Returns the separated source signals and separating matrix. 
%   Note in this algorithm, input mixed signals X should be preprocessed
%   by PCA, and there are no redundancy in it
%      
% Parameter ......
%            X     : mixed source signals in matix form N*Pwhere N represent how many 
%                    input(mixed) singals sources there are, and where P represent 
%                    how many samples. Note X should be preprocessed by PCA
% maxIteration     : how many iteration do you want
%     distance     : distance between new separating matrix and the old separating one when 
%                    separating matrix converging.
%        initB     : initial separating matrix B
%     
% Samples ......
%   B = rand(size(X,1));
%   [Y,B] = fastICAML(X,maxIteration',20,'distance',0.01,'initB',B,'report','on');
%         Compute the separating matrix B and separated source signals Y with
%         parameters: number of iterations is 20, initial separating matrix
%         B, and convergence distance 0.01. Still ,print some information
%         about progress of computing.
% 
% See also:
%   ffpMLdemo
%
% Author: Zhang Zhi-Lin
%         zhangzl@vip.163.com
% Version: 1.1
% Date: Nov.8,2003
%
% Reference:
%   [1] Aapo Hyvarinen, Juha Karhunen, Erkki Oja, Independent Component
%       Analysis. John Wiley & Sons.Inc, 2001. Page: 209-212
%


%=====================================================
% Default value
maxIteration = 50;
distance = 0.1; 
report = 'on';

N = size(X,1);
P = size(X,2);  

inputB = 0;  % tell whether initial B is input by user. If is 1, B is input by user 


%------------------------------------------------
% Get optional parameter
if(mod(length(varargin),2)==1)
    error('Optional parameters should always go by pairs\n');
else
    for i=1:2:(length(varargin)-1)
        switch lower(varargin{i})
            case 'maxiteration'
                maxIteration = varargin{i+1};
            case 'distance'
                distance = varargin{i+1}; 
            case 'initb'
                B = varargin{i+1}; 
                inputB = 1;   % initial B is input by user
            case 'report'
                report=varargin{i+1};  report=lower(report);
            otherwise
                % input name is wrong:
                error(['Unrecognized parameter: ''' varargin{i} '''']);
        end    % switch
    end   % for
end  % if

if inputB == 0
    % then initial separating matrix B is generated by computer
    B = rand(N)-0.5;
end

%------------------------------------------------
% Check if input parameter is correct
if ~( strcmp(report,'on') | strcmp(report,'off') )
    error(sprintf('Illegal value [ %s ] for parameter: ''report''\n', report));
end

if ~isreal(X)
    error('Input matrix has an imaginary part.\n');
end

if (~isnumeric(maxIteration) ) |  maxIteration < 1 
    error(sprintf('Illegal value for parameter: ''maxIteration''\n', maxIteration));
end

if ~isnumeric(distance) 
    error(sprintf('Illegal value for parameter: ''distance'', it must be number.\n', distance));
end

if ( size(B,1) ~= size(B,2) ) | ( ~isreal(B) ) | size(B,1) ~= N 
    error(sprintf('Illegal value for parameter: ''initB''\n'));
end
            
%------------------------------------------------
% print some infomation 
if strcmp(report,'on') 
    fprintf('\n=======================================\n');
    fprintf('   Information about parameters...\n');
    fprintf('=======================================\n');
    fprintf('Number of input signals: %d\n',size(X,1));
    fprintf('Length of signals:%d\n',P);    
    fprintf('Initial B = \n');disp(B);
    fprintf('Nonlinear function is tanh \n');
    fprintf('maxIteration = %d\n',maxIteration);
    fprintf('Converge conditiondistance <= %d\n',distance);
    fprintf('report = %s\n',report);
end

if strcmp(report,'on') 
    fprintf('\n=======================================\n');
    fprintf('  Starting FastICA calculation for ML...\n'); 
    fprintf('=======================================\n');
end


C = cov(X',1);     % compute correlation matrix C=E{XX'}

BOld=zeros(N);     % store the previous value of separating matrix

for loop = 1:maxIteration+1     
    % Convergence?
    if loop == maxIteration + 1
        fprintf('Cannot converge even after %d iteration\n',maxIteration);
        Y=B*X;
        return;
    end
    
    Y = B*X;
    
    % Compute the correct parameter for each iteration
    for k = 1:N   
        beta(k,1) = - 1/P * tanh(Y(k,:))*Y(k,:)';
        apha(k,1) = -1/( beta(k,1) + (1 - 1/P*tanh(Y(k,:))*tanh(Y(k,:))'  )  );
    end
    
    % Update separating matrix
    B = B + diag(apha)*( diag(beta) + 1/P*( tanh(Y)*Y' ) )*B;  
    
    % Decorrelate and normalize
    B = real(inv(B * C * B')^(1/2))*B;   
    
    % Test for termination condition.Here,the algorithm has converged if
    % the direction of B and BOld is the same or opposite.
    length = min( norm(B - BOld), norm(B + BOld) );    
    if( length <= distance) 
        fprintf('\n=======================================\n');
        fprintf('  Convergence after %d iteration\n',loop);
        fprintf('=======================================\n');
        Y=B*X;
        return; 
    else
        % Show the progress
        if strcmp(report,'on') 
            fprintf('Distance between new and old sep. matrix is %d\n',length);
            fprintf('Iteration No.%d\n--------------\n',loop);
        end
    end
    
    BOld=B;  

end




